/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __PIKA_IMAGE__UNDO_H__
#define __PIKA_IMAGE__UNDO_H__


gboolean        pika_image_undo_is_enabled      (PikaImage     *image);
gboolean        pika_image_undo_enable          (PikaImage     *image);
gboolean        pika_image_undo_disable         (PikaImage     *image);
gboolean        pika_image_undo_freeze          (PikaImage     *image);
gboolean        pika_image_undo_thaw            (PikaImage     *image);

gboolean        pika_image_undo                 (PikaImage     *image);
gboolean        pika_image_redo                 (PikaImage     *image);

gboolean        pika_image_strong_undo          (PikaImage     *image);
gboolean        pika_image_strong_redo          (PikaImage     *image);

PikaUndoStack * pika_image_get_undo_stack       (PikaImage     *image);
PikaUndoStack * pika_image_get_redo_stack       (PikaImage     *image);

void            pika_image_undo_free            (PikaImage     *image);

gint            pika_image_get_undo_group_count (PikaImage     *image);
gboolean        pika_image_undo_group_start     (PikaImage     *image,
                                                 PikaUndoType   undo_type,
                                                 const gchar   *name);
gboolean        pika_image_undo_group_end       (PikaImage     *image);

PikaUndo      * pika_image_undo_push            (PikaImage     *image,
                                                 GType          object_type,
                                                 PikaUndoType   undo_type,
                                                 const gchar   *name,
                                                 PikaDirtyMask  dirty_mask,
                                                 ...) G_GNUC_NULL_TERMINATED;

PikaUndo      * pika_image_undo_can_compress    (PikaImage     *image,
                                                 GType          object_type,
                                                 PikaUndoType   undo_type);


#endif /* __PIKA_IMAGE__UNDO_H__ */
